"""
Some threading related utilities.
"""

import threading

from typing import Callable, Optional, TypeVar, Generic

T = TypeVar("T")
class ThreadLocal(Generic[T]):
    """
    A thread local value holder
    """
    # constructor

    def __init__(self, default_factory: Optional[Callable[[], T]] = None):
        self.local = threading.local()
        self.factory = default_factory

    # public

    def get(self) -> Optional[T]:
        """
        return the current value or invoke the optional factory to compute one

        Returns:
            Optional[T]: the value associated with the current thread
        """
        if not hasattr(self.local, "value"):
            if self.factory is not None:
                self.local.value = self.factory()
            else:
                return None

        return self.local.value

    def set(self, value: T) -> None:
        """
        set a value in the current thread

        Args:
            value: the value
        """
        self.local.value = value

    def clear(self) -> None:
        """
        clear the value in the current thread
        """
        if hasattr(self.local, "value"):
            del self.local.value



"""
This module provides aspect-oriented programming (AOP) capabilities for Python applications.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import inspect
import re
import threading
import types
from dataclasses import dataclass
from enum import auto, Enum
from typing import Optional, Dict, Type, Callable

from aspyx.reflection import Decorators, TypeDescriptor
from aspyx.di import injectable, order, Providers, ClassInstanceProvider, Environment, PostProcessor

class AOPException(Exception):
    """
    Exception raised for errors in the aop logic.
    """

class AspectType(Enum):
    """
    AspectType defines the types of aspect-oriented advice that can be applied to methods.

    The available types are:

    - BEFORE: Advice to be executed before the method invocation.
    - AROUND: Advice that intercepts the method invocation.
    - AFTER: Advice to be executed after the method invocation, regardless of its outcome.
    - ERROR: Advice to be executed if the method invocation raises an exception.

    These types are used to categorize and apply aspect logic at different points in a method's execution lifecycle.
    """
    BEFORE = auto()
    AROUND = auto()
    AFTER = auto()
    ERROR = auto()

class AspectTarget(ABC):
    """
    AspectTarget defines the target for an aspect. It can be used to specify the class, method, and conditions under which the aspect should be applied.
    It supports matching by class type, method name, patterns, decorators, and more.
    """

    # properties

    __slots__ = [
        "_function",
        "_type",
        "_async",
        "_clazz",
        "_instance",
        "names",
        "patterns",
        "types",
        "other",
        "decorators",
    ]

    # constructor

    def __init__(self):
        self._clazz = None
        self._instance = None
        self._async = False
        self._function = None
        self._type = None

        self.patterns = []
        self.names = []
        self.types = []
        self.decorators = []

        self.other : list[AspectTarget] = []

    # abstract

    def _matches(self, clazz : Type, func):
        if not self._matches_self(clazz, func):
            for target in self.other:
                if target._matches(clazz, func):
                    return True

            return False

        return True

    @abstractmethod
    def _matches_self(self, clazz: Type, func):
        pass

    # protected

    def _add(self, target: AspectTarget):
        self.other.append(target)
        return self

     # fluent

    def function(self, func) -> AspectTarget:
        self._function = func
        return self

    def type(self, type: AspectType):
        self._type = type

        return self

    def that_are_async(self) -> AspectTarget:
        """
        matches methods that are async

        Returns:
            AspectTarget: self
        """
        self._async = True
        return self

    def of_type(self, type: Type) -> AspectTarget:
        """
        matches methods belonging to a class or classes that are subclasses of the specified type

        Args:
            type (Type): the type to match against

        Returns:
            AspectTarget: self
        """
        self.types.append(type)
        return self

    def decorated_with(self, decorator: Callable) -> AspectTarget:
        """
        matches methods or classes that are decorated with the specified decorator

        Args:
            decorator (Callable): the decorator callable

        Returns:
            AspectTarget: self
        """
        self.decorators.append(decorator)
        return self

    def matches(self, pattern: str) -> AspectTarget:
        """
        Matches the target against a pattern.

        Args:
            pattern (str): the pattern

        Returns:
            AspectTarget: self
        """
        self.patterns.append(re.compile(pattern))
        return self

    def named(self, name: str) -> AspectTarget:
        """
        Matches the target against a name.

        Args:
            name (str): the name

        Returns:
            AspectTarget: self
        """
        self.names.append(name)
        return self

class ClassAspectTarget(AspectTarget):
    """
    An AspectTarget matching classes
    """
    # properties

    __slots__ = [
    ]

    # public

    def _matches_self(self, clazz : Type, func):
        class_descriptor = TypeDescriptor.for_type(clazz)
        #descriptor = TypeDescriptor.for_type(func)
        # type

        if len(self.types) > 0:
            if next((type for type in self.types if issubclass(clazz, type)), None) is None:
                return False

        # decorators

        if len(self.decorators) > 0:
            if next((decorator for decorator in self.decorators if class_descriptor.has_decorator(decorator)), None) is None:
                return False

        # names

        if len(self.names) > 0:
            if next((name for name in self.names if name == clazz.__name__), None) is None:
                return False

        # patterns

        if len(self.patterns) > 0:
            if next((pattern for pattern in self.patterns if re.fullmatch(pattern, clazz.__name__) is not None), None) is None:
                return False

        return True

    # fluent

class MethodAspectTarget(AspectTarget):
    """
       An AspectTarget matching methods
       """

    # properties

    __slots__ = [ ]

    # public

    def _matches_self(self, clazz : Type, func):
        descriptor = TypeDescriptor.for_type(clazz)

        method_descriptor = descriptor.get_method(func.__name__)

        # async

        if self._async is not method_descriptor.is_async():
            return False

        # type

        if len(self.types) > 0:
            if next((type for type in self.types if issubclass(clazz, type)), None) is None:
                return False

        # decorators

        if len(self.decorators) > 0:
            if next((decorator for decorator in self.decorators if method_descriptor.has_decorator(decorator)), None) is None:
                return False

        # names

        if len(self.names) > 0:
            if next((name for name in self.names if name == func.__name__), None) is None:
                return False

        # patterns

        if len(self.patterns) > 0:
            if next((pattern for pattern in self.patterns if re.fullmatch(pattern, func.__name__) is not None), None) is None:
                return False

        # yipee

        return True

def methods() -> AspectTarget:
    """
    Create a new AspectTarget instance to define method aspect targets.

    Returns:
         AspectTarget: the method target
    """
    return MethodAspectTarget()

def classes() -> AspectTarget:
    """
    Create a new AspectTarget instance to define class aspect targets.

    Returns:
        AspectTarget: the method target
    """
    return ClassAspectTarget()


class Aspect:
    __slots__ = [
        "next",
    ]

    # constructor

    def __init__(self, next: 'Aspect'):
        self.next = next

    # public

    def call(self, invocation: 'Invocation'):
        pass

    async def call_async(self, invocation: 'Invocation'):
        pass

class FunctionAspect(Aspect):
    __slots__ = [
        "instance",
        "func",
        "order",
    ]

    def __init__(self, instance, func, next_aspect: Optional['Aspect']):
        super().__init__(next_aspect)

        self.instance = instance
        self.func = func

        self.order = next((decorator.args[0] for decorator in Decorators.get(func) if decorator.decorator is order), 0)

    def call(self, invocation: 'Invocation'):
        invocation.current_aspect = self

        return self.func(self.instance, invocation)

    async def call_async(self, invocation: 'Invocation'):
        invocation.current_aspect = self

        return await self.func(self.instance, invocation)

class MethodAspect(FunctionAspect):
    __slots__ = []

    def __init__(self, instance, func):
        super().__init__(instance, func, None)

    def call(self, invocation: 'Invocation'):
        invocation.current_aspect = self

        return self.func(*invocation.args, **invocation.kwargs)

    async def call_async(self, invocation: 'Invocation'):
        invocation.current_aspect = self

        return await self.func(*invocation.args, **invocation.kwargs)

@dataclass
class Aspects:
    before: list[Aspect]
    around: list[Aspect]
    error: list[Aspect]
    after: list[Aspect]

class Invocation:
    """
    Invocation stores the relevant data of a single method invocation.
    It holds the arguments, keyword arguments, result, error, and the aspects that define the aspect behavior.
    """
    # properties

    __slots__ = [
        "func",
        "args",
        "kwargs",
        "result",
        "exception",
        "aspects",
        "current_aspect",
    ]

    # constructor

    def __init__(self, func, aspects: Aspects):
        self.func = func
        self.args : list[object] = []
        self.kwargs = None
        self.result = None
        self.exception = None
        self.aspects = aspects
        self.current_aspect = None

    def call(self, *args, **kwargs):
        # remember args

        self.args = args
        self.kwargs = kwargs

        # run all before

        for aspect in self.aspects.before:
            aspect.call(self)

        # run around's with the method being the last aspect!

        try:
            self.result = self.aspects.around[0].call(self) # will follow the proceed chain

        except Exception as e:
            self.exception = e
            for aspect in self.aspects.error:
                aspect.call(self)

        # run all before

        for aspect in self.aspects.after:
            aspect.call(self)

        if self.exception is not None:
            raise self.exception # rethrow the error

        return self.result

    async def call_async(self, *args, **kwargs):
        # remember args

        self.args = args
        self.kwargs = kwargs

        # run all before

        for aspect in self.aspects.before:
            aspect.call(self)

        # run around's with the method being the last aspect!

        try:
            self.result = await self.aspects.around[0].call_async(self) # will follow the proceed chain

        except Exception as e:
            self.exception = e
            for aspect in self.aspects.error:
                aspect.call(self)

        # run all before

        for aspect in self.aspects.after:
            aspect.call(self)

        if self.exception is not None:
            raise self.exception # rethrow the error

        return self.result

    def proceed(self, *args, **kwargs):
        """
        Proceed to the next aspect in the around chain up to the original method.
        """
        if len(args) > 0 or len(kwargs) > 0:  # as soon as we have args, we replace the current ones
            self.args = args
            self.kwargs = kwargs

        # next one please...

        return self.current_aspect.next.call(self)

    async def proceed_async(self, *args, **kwargs):
        """
        Proceed to the next aspect in the around chain up to the original method.
        """
        if len(args) > 0 or len(kwargs) > 0:  # as soon as we have args, we replace the current ones
            self.args = args
            self.kwargs = kwargs

        # next one please...

        return await self.current_aspect.next.call_async(self)

class Advices:
    """
    Internal utility class that collects all advice s
    """
    # static data

    targets: list[AspectTarget] = []

    __slots__ = []

    # constructor

    def __init__(self):
        pass

    # methods

    @classmethod
    def collect(cls, clazz, member, type: AspectType, environment: Environment):
        aspects = [
            FunctionAspect(environment.get(target._clazz), target._function, None) for target in Advices.targets
            if target._type == type
               and target._clazz is not clazz
               and environment.providers.get(target._clazz) is not None
               and target._matches(clazz, member)
        ]

        # sort according to order

        aspects = sorted(aspects, key=lambda aspect: aspect.order)

        # link

        for i in range(0, len(aspects) - 1):
            aspects[i].next = aspects[i + 1]

        # done

        return aspects

    @classmethod
    def aspects_for(cls, instance, environment: Environment) -> Dict[Callable,Aspects]:
        clazz = type(instance)

        result = {}

        for _, member in inspect.getmembers(clazz, predicate=inspect.isfunction):
            aspects = cls.compute_aspects(clazz, member, environment)
            if aspects is not None:
                result[member] = aspects

        # add around methods

        value = {}

        for key, cjp in result.items():
            jp = Aspects(
                before=cjp.before,
                around=cjp.around,
                error=cjp.error,
                after=cjp.after)

            # add method to around

            jp.around.append(MethodAspect(instance, key))
            if len(jp.around) > 1:
                jp.around[len(jp.around) - 2].next = jp.around[len(jp.around) - 1]

            value[key] = jp

        # done

        return value

    @classmethod
    def compute_aspects(cls, clazz, member, environment: Environment) -> Optional[Aspects]:
        befores = cls.collect(clazz, member, AspectType.BEFORE, environment)
        arounds = cls.collect(clazz, member, AspectType.AROUND, environment)
        afters = cls.collect(clazz, member, AspectType.AFTER, environment)
        errors = cls.collect(clazz, member, AspectType.ERROR, environment)

        if len(befores) > 0 or len(arounds) > 0 or len(afters) > 0  or len(errors) > 0:
            return Aspects(
                before=befores,
                around=arounds,
                error=errors,
                after=afters
            )
        else:
            return None

def sanity_check(clazz: Type, name: str):
    m = TypeDescriptor.for_type(clazz).get_method(name)
    if len(m.param_types) != 1 or m.param_types[0] != Invocation:
        raise AOPException(f"Method {clazz.__name__}.{name} expected to have one parameter of type Invocation")

# decorators

def advice(cls):
    """
    Classes decorated with `@advice` are treated as advice classes.
    They can contain methods decorated with `@before`, `@after`, `@around`, or `@error` to define aspects.
    """
    Providers.register(ClassInstanceProvider(cls, True))

    Decorators.add(cls, advice)

    for name, member in TypeDescriptor.for_type(cls).methods.items():
        decorator = next((decorator for decorator in member.decorators if decorator.decorator in [before, after, around, error]), None)
        if decorator is not None:
            target = decorator.args[0] # multiple targets are already merged in a single! check _register
            target._clazz = cls
            sanity_check(cls, name)
            Advices.targets.append(target) #??

    return cls


# decorators

def _register(decorator, targets: list[AspectTarget], func, aspect_type: AspectType):
    target = targets[0]

    for i in range(1, len(targets)):
        target._add(targets[i])

    target.function(func).type(aspect_type)

    Decorators.add(func, decorator, target)

def before(*targets: AspectTarget):
    """
    Methods decorated with `@before` will be executed before the target method is invoked.
    """
    def decorator(func):
        _register(before, targets, func, AspectType.BEFORE)

        return func

    return decorator

def error(*targets: AspectTarget):
    """
    Methods decorated with `@error` will be executed if the target method raises an exception."""
    def decorator(func):
        _register(error, targets, func, AspectType.ERROR)

        return func

    return decorator

def after(*targets: AspectTarget):
    """
    Methods decorated with `@after` will be executed after the target method is invoked.
    """
    def decorator(func):
        _register(after, targets, func, AspectType.AFTER)

        return func

    return decorator

def around(*targets: AspectTarget):
    """
    Methods decorated with `@around` will be executed around the target method.
    Every around method must accept a single parameter of type Invocation and needs to call proceed
    on this parameter to proceed to the next around method.
    """
    def decorator(func):
        _register(around, targets, func, AspectType.AROUND)

        return func

    return decorator

@injectable(scope="environment")
@order(0)
class AdviceProcessor(PostProcessor):
    # properties

    __slots__ = [
        "lock",
        "cache"
    ]

    # constructor

    def __init__(self):
        super().__init__()

        self.cache : Dict[Type, Dict[Callable,Aspects]] = {}
        self.lock = threading.RLock()

    # local

    def aspects_for(self, instance, environment: Environment) -> Dict[Callable,Aspects]:
        clazz = type(instance)
        result = self.cache.get(clazz, None)
        if result is None:
            with self.lock:
                result = Advices.aspects_for(instance, environment)
                self.cache[clazz] = result # TOID der cache ist zu dick?????

        return result

    # implement

    def process(self, instance: object, environment: Environment):
        aspect_dict = self.aspects_for(instance, environment)

        for member, aspects in aspect_dict.items():
            Environment.logger.debug("add aspects for %s:%s", type(instance), member.__name__)

            def wrap(jp):
                def sync_wrapper(*args, **kwargs):
                    return Invocation(member, jp).call(*args, **kwargs)

                return sync_wrapper

            def wrap_async(jp):
                async def async_wrapper(*args, **kwargs):
                    return await Invocation(member, jp).call_async(*args, **kwargs)

                return async_wrapper

            if inspect.iscoroutinefunction(member):
                setattr(instance, member.__name__, types.MethodType(wrap_async(aspects), instance))
            else:
                setattr(instance, member.__name__, types.MethodType(wrap(aspects), instance))



"""
Tests for the AOP (Aspect-Oriented Programming) functionality in the aspyx.di module.
"""
from __future__ import annotations

import asyncio
import logging
import threading
import unittest
from typing import Dict

from aspyx.di.threading import synchronized
from aspyx.reflection import Decorators
from aspyx.di import injectable, Environment, module, order
from aspyx.di.aop import advice, before, after, around, methods, Invocation, error, classes

logging.basicConfig(
    level=logging.DEBUG,
    format='[%(asctime)s] %(levelname)s in %(filename)s:%(lineno)d - %(message)s'
)

def configure_logging(levels: Dict[str, int]) -> None:
    for name in levels:
        logging.getLogger(name).setLevel(levels[name])

configure_logging({"aspyx": logging.DEBUG})

def transactional():
    def decorator(func):
        Decorators.add(func, transactional)
        return func #

    return decorator

@module()
class SampleModule:
    def __init__(self):
        pass


@injectable()
@transactional()
class Bar:
    def __init__(self):
        pass

    async def say_async(self, message: str):
        await asyncio.sleep(0.01)

        return f"hello {message}"

    @synchronized()
    def say(self, message: str):
        return f"hello {message}"

@injectable(eager=False, scope="request")
class Foo:
    def __init__(self, bar: Bar):
        self.bar = bar

    @synchronized()
    def say(self, message: str):
        return f"hello {message}"

    def throw_error(self):
        raise Exception("ouch")

@advice
class SampleAdvice:
    # constructor

    def __init__(self):
        self.name = "SampleAdvice"

        self.before_calls = 0
        self.after_calls = 0
        self.around_calls = 0
        self.error_calls = 0
        self.afters = []

        self.exception = None

    # public

    def reset(self):
        self.before_calls = 0
        self.after_calls = 0
        self.around_calls = 0
        self.error_calls = 0

        self.afters.clear()

        self.exception = None

    # aspects

    @error(methods().of_type(Foo).matches(".*"))
    def error(self, invocation: Invocation):
        self.exception = invocation.exception

    @before(methods().named("say").of_type(Foo).matches(".*"))
    def call_before_foo(self, invocation: Invocation):
        self.before_calls += 1

    @before(methods().named("say").of_type(Foo).matches(".*"))
    @order(10)
    def other_before_foo(self, invocation: Invocation):
        pass

    @before(methods().named("say").of_type(Bar))
    def call_before_bar(self, invocation: Invocation):
        self.before_calls += 1

    @after(methods().named("say"))
    def call_after(self, invocation: Invocation):
        self.after_calls += 1
        self.afters.append(0)

    @after(methods().named("say"))
    @order(2)
    def call_after2(self, invocation: Invocation):
        self.after_calls += 1
        self.afters.append(2)

    @after(methods().named("say"))
    @order(1)
    def call_after1(self, invocation: Invocation):
        self.after_calls += 1
        self.afters.append(1)

    @around(methods().that_are_async())
    async def call_around_async(self, invocation: Invocation):
        self.around_calls += 1

        print("call_around_async")

        return await invocation.proceed_async()

    @around(methods().named("say"))
    def call_around(self, invocation: Invocation):
        self.around_calls += 1

        args = [invocation.args[0],invocation.args[1] + "!"]

        return invocation.proceed(*args)

    @around(methods().decorated_with(transactional), classes().decorated_with(transactional))
    def call_transactional1(self, invocation: Invocation):
        self.around_calls += 1

        return invocation.proceed()

    #@around(classes().decoratedWith(transactional))
    def call_transactional(self, invocation: Invocation):
        self.around_calls += 1

        return invocation.proceed()

environment = Environment(SampleModule)

class TestAsyncAdvice(unittest.IsolatedAsyncioTestCase):

    def test_thread_test(self):
        n_threads = 1
        iterations = 10000

        threads = []

        def worker(thread_id: int):
            env =  Environment(SampleModule)

            for i in range(iterations):
                foo = env.get(Foo)

                foo.say(f"thread {thread_id}")

        for t_id in range(0, n_threads):
            thread = threading.Thread(target=worker, args=(t_id,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        print("All threads finished.")

    async def test_async(self):
        bar = environment.get(Bar)

        result = await bar.say_async("world")

        self.assertEqual(result, "hello world")

class TestAdvice(unittest.TestCase):
    def test_order(self):
        advice = environment.get(SampleAdvice)

        advice.reset()
        foo = environment.get(Foo)

        foo.say("world")

        self.assertTrue( advice.afters == sorted( advice.afters))

    def test_advice(self):
        advice = environment.get(SampleAdvice)

        advice.reset()
        foo = environment.get(Foo)

        self.assertIsNotNone(foo)

        # foo

        result = foo.say("world")

        self.assertEqual(result, "hello world!")

        self.assertEqual(advice.before_calls, 1)
        self.assertEqual(advice.around_calls, 1)
        self.assertEqual(advice.after_calls, 3)

        advice.reset()

        # bar

        result = foo.bar.say("world")

        self.assertEqual(result, "hello world!")

        self.assertEqual(advice.before_calls, 1)
        self.assertEqual(advice.around_calls, 2)
        self.assertEqual(advice.after_calls, 3)

    def test_error(self):
        foo = environment.get(Foo)
        advice = environment.get(SampleAdvice)

        advice.reset()

        try:
            foo.throw_error()
        except Exception as e:#
            self.assertIs(e, advice.exception)

        # foo

        foo.say("hello")

if __name__ == '__main__':
    unittest.main()



"""
This module provides dependency injection and aop capabilities for Python applications.
"""
from .di import InstanceProvider, conditional, requires_class, requires_feature, DIException, AbstractCallableProcessor, LifecycleCallable, Lifecycle, Providers, Environment, ClassInstanceProvider, injectable, factory, module, inject, order, create, on_init, on_running, on_destroy, inject_environment, Factory, PostProcessor

# import something from the subpackages, so that the decorators are executed

from .configuration import ConfigurationManager
from .aop import before
from .threading import SynchronizeAdvice

imports = [ConfigurationManager, before, SynchronizeAdvice]

__all__ = [
    "ClassInstanceProvider",
    "Providers",
    "Environment",
    "injectable",
    "factory",
    "module",
    "inject",
    "create",
    "order",
    "InstanceProvider",
    "on_init",
    "on_running",
    "on_destroy",
    "inject_environment",
    "Factory",
    "PostProcessor",
    "AbstractCallableProcessor",
    "LifecycleCallable",
    "DIException",
    "Lifecycle",
    "conditional",
    "requires_class",
    "requires_feature"
]



"""
Test cases for the TypeDescriptor and Decorators functionality in aspyx.reflection.
"""
from __future__ import annotations

import unittest

from aspyx.reflection import TypeDescriptor, Decorators


def transactional():
    def decorator(func):
        Decorators.add(func, transactional)
        return func #

    return decorator

@transactional()
class Base:
    def __init__(self):
        pass

    @transactional()
    def base(self, message: str) -> str:
        pass

    def no_type_hints(self, message):
        pass

class Derived(Base):
    @classmethod
    def foo(cls):
        pass

    def derived(self, message: str) -> str:
        pass

class TestReflection(unittest.TestCase):
    def test_decorators(self):
        base_descriptor = TypeDescriptor.for_type(Base)

        self.assertTrue(base_descriptor.has_decorator(transactional))
        self.assertTrue( base_descriptor.get_method("base").has_decorator(transactional))

    def test_methods(self):
        derived_descriptor = TypeDescriptor.for_type(Derived)

        self.assertIsNotNone(derived_descriptor.get_method("derived").return_type, str)


if __name__ == '__main__':
    unittest.main()



"""
Test for DynamicProxy in aspyx.reflection
"""
from __future__ import annotations

import unittest

from aspyx.reflection import DynamicProxy


class Handler(DynamicProxy.InvocationHandler):
    def invoke(self, invocation: DynamicProxy.Invocation):
        return invocation.args[0]

    async def invoke_async(self, invocation: 'DynamicProxy.Invocation'):
        return invocation.args[0]

class Service:
    def say(self, message: str) -> str:
        pass

    async def say_async(self, message: str) -> str:
        return message

class TestAsyncProxyAsync(unittest.IsolatedAsyncioTestCase):
    async def test_async_proxy(self):
        proxy = DynamicProxy.create(Service, Handler())

        answer = await proxy.say_async("hello")
        self.assertEqual(answer, "hello")

class TestProxy(unittest.TestCase):
    def test_proxy(self):
        proxy = DynamicProxy.create(Service, Handler())

        answer = proxy.say("hello")
        self.assertEqual(answer, "hello")


if __name__ == '__main__':
    unittest.main()



"""
This module provides a TypeDescriptor class that allows introspection of Python classes,
including their methods, decorators, and type hints. It supports caching for performance
"""
from __future__ import annotations

import inspect
from inspect import signature
import threading
from types import FunctionType

from typing import Callable, get_type_hints, Type, Dict, Optional
from weakref import WeakKeyDictionary


class DecoratorDescriptor:
    """
    A DecoratorDescriptor covers the decorator - a callable - and the passed arguments
    """
    __slots__ = [
        "decorator",
        "args"
    ]

    def __init__(self, decorator: Callable, *args):
        self.decorator = decorator
        self.args = args

    def __str__(self):
        return f"@{self.decorator.__name__}({', '.join(map(str, self.args))})"

class Decorators:
    """
    Utility class that caches decorators ( Python does not have a feature for this )
    """
    @classmethod
    def add(cls, func_or_class, decorator: Callable, *args):
        """
        Remember the decorator
        Args:
            func_or_class: a function or class
            decorator: the decorator
            *args: any arguments supplied to the decorator
        """
        decorators = getattr(func_or_class, '__decorators__', None)
        if decorators is None:
            setattr(func_or_class, '__decorators__', [DecoratorDescriptor(decorator, *args)])
        else:
            decorators.append(DecoratorDescriptor(decorator, *args))

    @classmethod
    def has_decorator(cls, func_or_class, callable: Callable) -> bool:
        """
        Return True, if the function or class is decorated with the decorator
        Args:
            func_or_class: a function or class
            callable: the decorator

        Returns:
            bool: the result
        """
        return any(decorator.decorator is callable for decorator in Decorators.get(func_or_class))

    @classmethod
    def get_decorator(cls, func_or_class, callable: Callable) -> DecoratorDescriptor:
        return next((decorator for decorator in Decorators.get_all(func_or_class) if decorator.decorator is callable), None)

    @classmethod
    def get_all(cls, func_or_class) -> list[DecoratorDescriptor]:
        return getattr(func_or_class, '__decorators__', [])

    @classmethod
    def get(cls, func_or_class) -> list[DecoratorDescriptor]:
        """
        return the list of decorators associated with the given function or class
        Args:
            func_or_class: the function or class

        Returns:
            list[DecoratorDescriptor]: the list
        """
        if inspect.ismethod(func_or_class):
            func_or_class = func_or_class.__func__  # unwrap bound method

        #return getattr(func_or_class, '__decorators__', []) will return inherited as well
        return func_or_class.__dict__.get('__decorators__', [])


class TypeDescriptor:
    """
    This class provides a way to introspect Python classes, their methods, decorators, and type hints.
    """
    # inner class

    class MethodDescriptor:
        """
        This class represents a method of a class, including its decorators, parameter types, and return type.
        """
        # constructor

        def __init__(self, cls, method: Callable):
            self.clazz = cls
            self.method = method
            self.decorators: list[DecoratorDescriptor] = Decorators.get(method)
            self.param_types : list[Type] = []

            type_hints = get_type_hints(method)
            sig = signature(method)

            for name, _ in sig.parameters.items():
                if name != 'self':
                    self.param_types.append(type_hints.get(name, object))

            self.return_type = type_hints.get('return', None)

        # public

        def get_name(self) -> str:
            """
            return the method name

            Returns:
                str: the method name
            """
            return self.method.__name__

        def get_doc(self, default = "") -> str:
            """
            return the method docstring

            Args:
                default: the default if no docstring is found

            Returns:
                str: the docstring
            """
            return self.method.__doc__ or default

        def is_async(self) -> bool:
            """
            return true if the method is asynchronous

            Returns:
                bool: async flag
            """
            return inspect.iscoroutinefunction(self.method)

        def get_decorators(self) -> list[DecoratorDescriptor]:
            return self.decorators

        def get_decorator(self, decorator: Callable) -> Optional[DecoratorDescriptor]:
            """
            return the DecoratorDescriptor - if any - associated with the passed Callable

            Args:
                decorator: the decorator

            Returns:
                Optional[DecoratorDescriptor]: the DecoratorDescriptor or None
            """
            for dec in self.decorators:
                if dec.decorator is decorator:
                    return dec

            return None

        def has_decorator(self, decorator: Callable) -> bool:
            """
            return True if the method is decorated with the decorator

            Args:
                decorator: the decorator callable

            Returns:
                bool: True if the method is decorated with the decorator
            """
            for dec in self.decorators:
                if dec.decorator is decorator:
                    return True

            return False

        def __str__(self):
            return f"Method({self.method.__name__})"

    # class properties

    _cache = WeakKeyDictionary()
    _lock = threading.RLock()

    # class methods

    @classmethod
    def for_type(cls, clazz: Type) -> TypeDescriptor:
        """
        Returns a TypeDescriptor for the given class, using a cache to avoid redundant introspection.
        """
        descriptor = cls._cache.get(clazz)
        if descriptor is None:
            with cls._lock:
                descriptor = cls._cache.get(clazz)
                if descriptor is None:
                    descriptor = TypeDescriptor(clazz)
                    cls._cache[clazz] = descriptor

        return descriptor

    # constructor

    def __init__(self, cls):
        self.cls = cls
        self.decorators = Decorators.get(cls)
        self.methods: Dict[str, TypeDescriptor.MethodDescriptor] = {}
        self.local_methods: Dict[str, TypeDescriptor.MethodDescriptor] = {}

        # check superclasses

        self.super_types = [TypeDescriptor.for_type(x) for x in cls.__bases__ if not x is object]

        for super_type in self.super_types:
            self.methods = self.methods | super_type.methods

        # methods

        for name, member in self._get_local_members(cls):
            method = TypeDescriptor.MethodDescriptor(cls, member)
            self.local_methods[name] = method
            self.methods[name] = method

    # internal

    def _get_local_members(self, cls):
        #return [
        #    (name, value)
        #    for name, value in getmembers(cls, predicate=inspect.isfunction)
        #    if name in cls.__dict__
        #]

        return [
            (name, attr)
            for name, attr in cls.__dict__.items()
            if isinstance(attr, FunctionType)
        ]

    # public

    def get_decorator(self, decorator: Callable) -> Optional[DecoratorDescriptor]:
        """
        Returns the first decorator of the given type, or None if not found.
        """
        for dec in self.decorators:
            if dec.decorator is decorator:
                return dec

        return None

    def has_decorator(self, decorator: Callable) -> bool:
        """
        Checks if the class has a decorator of the given type."""
        for dec in self.decorators:
            if dec.decorator is decorator:
                return True

        return False

    def get_methods(self, local = False) ->  list[TypeDescriptor.MethodDescriptor]:
        """
        Returns a list of MethodDescriptor objects for the class.
        If local is True, only returns methods defined in the class itself, otherwise includes inherited methods.
        """
        if local:
            return list(self.local_methods.values())
        else:
            return list(self.methods.values())

    def get_method(self, name: str, local = False) -> Optional[TypeDescriptor.MethodDescriptor]:
        """
        Returns a MethodDescriptor for the method with the given name.
        If local is True, only searches for methods defined in the class itself, otherwise includes inherited methods.
        """
        if local:
            return self.local_methods.get(name, None)
        else:
            return self.methods.get(name, None)



"""
This module provides exception handling functions.
"""
from .exception_manager import exception_handler, handle, ExceptionManager

__all__ = [
    "exception_handler",
    "handle",
    "ExceptionManager"
]



"""
Configuration handling module.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Type, TypeVar

from aspyx.di import injectable, Environment, LifecycleCallable, Lifecycle
from aspyx.di.di import order, inject
from aspyx.reflection import Decorators, DecoratorDescriptor, TypeDescriptor

T = TypeVar("T")

class ConfigurationException(Exception):
    """
    Exception raised for errors in the configuration logic.
    """

@injectable()
class ConfigurationManager:
    """
    ConfigurationManager is responsible for managing different configuration sources by merging the different values
    and offering a uniform api.
    """

    __slots__ = ["sources", "_data", "coercions"]

    # constructor

    def __init__(self):
        self.sources = []
        self._data = {}
        self.coercions = {
            int: int,
            float: float,
            bool: lambda v: str(v).lower() in ("1", "true", "yes", "on"),
            str: str,
            # Add more types as needed
        }

    # internal

    def _register(self, source: ConfigurationSource):
        self.sources.append(source)
        self.load_source(source)

    # public

    def load_source(self,  source: ConfigurationSource):
        def merge_dicts(a: dict, b: dict) -> dict:
            result = a.copy()
            for key, b_val in b.items():
                if key in result:
                    a_val = result[key]
                    if isinstance(a_val, dict) and isinstance(b_val, dict):
                        result[key] = merge_dicts(a_val, b_val)  # Recurse
                    else:
                        result[key] = b_val  # Overwrite
                else:
                    result[key] = b_val
            return result

        self._data = merge_dicts(self._data, source.load())

    def get(self, path: str, type: Type[T], default : Optional[T]=None) -> T:
        """
        Retrieve a configuration value by path and type, with optional coercion.

        Args:
            path (str): The path to the configuration value, e.g. "database.host".
            type (Type[T]): The expected type.
            default (Optional[T]): The default value to return if the path is not found.

        Returns:
            T: The configuration value coerced to the specified type, or the default value if not found.
        """
        def resolve_value(path: str, default=None) -> T:
            keys = path.split(".")
            current = self._data
            for key in keys:
                if not isinstance(current, dict) or key not in current:
                    return default
                current = current[key]

            return current

        v = resolve_value(path, default)

        if isinstance(v, type):
            return v

        if type in self.coercions:
            try:
                return self.coercions[type](v)
            except Exception as e:
                raise ConfigurationException(f"error during coercion to {type}") from e
        else:
            raise ConfigurationException(f"unknown coercion to {type}")


class ConfigurationSource(ABC):
    """
    A configuration source is a provider of configuration data.
    """

    __slots__ = []

    def __init__(self):
        pass

    @inject()
    def set_manager(self, manager: ConfigurationManager):
        manager._register(self)

    @abstractmethod
    def load(self) -> dict:
        """
        return the configuration values of this source as a dictionary.
        """

# decorator

def inject_value(key: str, default=None):
    """
    Decorator to inject a configuration value into a method.

    Args:
        key (str): The configuration key to inject.
        default: The default value to use if the key is not found.

    """
    def decorator(func):
        Decorators.add(func, inject_value, key, default)

        return func

    return decorator

@injectable()
@order(9)
class ConfigurationLifecycleCallable(LifecycleCallable):
    def __init__(self,  manager: ConfigurationManager):
        super().__init__(inject_value, Lifecycle.ON_INJECT)

        self.manager = manager

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [self.manager.get(decorator.args[0], method.param_types[0], decorator.args[1])]



"""
Dynamic proxies for method interception and delegation.
"""
import inspect
from typing import Generic, TypeVar, Type, Callable

T = TypeVar("T")

class DynamicProxy(Generic[T]):
    """
    DynamicProxy enables dynamic method interception and delegation for any class type.

    It is used to create proxy objects that forward method calls to a custom InvocationHandler.
    This allows for advanced patterns such as aspect-oriented programming, logging, or remote invocation,
    by intercepting method calls at runtime and handling them as needed.

    Usage:
    ```python
        class MyHandler(DynamicProxy.InvocationHandler):
            def invoke(self, invocation):
                print(f"Intercepted: {invocation.name}")
                # custom logic here
                return ...

        proxy = DynamicProxy.create(SomeClass, MyHandler())
        proxy.some_method(args)  # Will be intercepted by MyHandler.invoke
    ```
    """
    # inner class

    class Invocation:
        __slots__ = [
            "type",
            "method",
            "args",
            "kwargs",
        ]

        # constructor

        def __init__(self, type: Type[T], method: Callable, *args, **kwargs):
            self.type = type
            self.method = method
            self.args = args
            self.kwargs = kwargs

    class InvocationHandler:
        def invoke(self, invocation: 'DynamicProxy.Invocation'):
            pass

        async def invoke_async(self, invocation: 'DynamicProxy.Invocation'):
            return self.invoke(invocation)

    # class methods

    @classmethod
    def create(cls, type: Type[T], invocation_handler: 'DynamicProxy.InvocationHandler') -> T:
        return DynamicProxy(type, invocation_handler)

    __slots__ = [
        "type",
        "invocation_handler"
    ]

    # constructor

    def __init__(self, type: Type[T], invocation_handler: 'DynamicProxy.InvocationHandler'):
        self.type = type
        self.invocation_handler = invocation_handler

    # public

    def __getattr__(self, name):
        method = getattr(self.type, name)

        if  inspect.iscoroutinefunction(method):
            async def async_wrapper(*args, **kwargs):
                return  await self.invocation_handler.invoke_async(DynamicProxy.Invocation(self.type, method, *args, **kwargs))

            return async_wrapper

        else:
            def sync_wrapper(*args, **kwargs):
                return self.invocation_handler.invoke(DynamicProxy.Invocation(self.type, method, *args, **kwargs))

            return sync_wrapper



"""
Test cases for the Configuration system in aspyx.di
"""
from __future__ import annotations

import unittest

from aspyx.di import injectable, Environment, module, create
from aspyx.di.aop import Invocation, advice, error, methods
from aspyx.exception import ExceptionManager, exception_handler, handle

class DerivedException(Exception):
    def __init__(self):
        pass

@module()
class SampleModule:
    # constructor

    def __init__(self):
        pass

    @create()
    def create_exception_manager(self) -> ExceptionManager:
        return ExceptionManager()

@injectable()
@exception_handler()
class Test:
    def __init__(self):
        pass

    @handle()
    def handle_derived_exception(self, exception: DerivedException):
        return ExceptionManager.proceed()

    @handle()
    def handle_exception(self, exception: Exception):
        return exception

    @handle()
    def handle_base_exception(self, exception: BaseException):
        return exception

@injectable()
class Service:
    def __init__(self):
        pass

    def throw(self):
        raise DerivedException()

@advice
class ExceptionAdvice:
    def __init__(self, exception_manager: ExceptionManager):
        self.exception_manager = exception_manager

    @error(methods().of_type(Service))
    def handle_error(self, invocation: Invocation):
        exception = self.exception_manager.handle(invocation.exception) # possibly transform

        invocation.exception = exception

class TestExceptionManager(unittest.TestCase):
    def test_exception_manager(self):
        environment =  Environment(SampleModule)

        service = environment.get(Service)

        try:
            service.throw()
        except Exception as e:
            pass

if __name__ == '__main__':
    unittest.main()



"""
Test cases for the aspyx.di module.
"""
from __future__ import annotations

import threading
import time
import logging
import unittest
from typing import Dict

from aspyx.di.configuration import EnvConfigurationSource

# not here

logging.basicConfig(
    level=logging.DEBUG,
    format='[%(asctime)s] %(levelname)s in %(filename)s:%(lineno)d - %(message)s'
)

def configure_logging(levels: Dict[str, int]) -> None:
    for name, level in levels.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)

        # If no handler is attached, add one
        if not logger.handlers and logger.propagate is False:
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(
                '[%(asctime)s] %(levelname)s in %(filename)s:%(lineno)d - %(message)s'
            ))
            logger.addHandler(handler)

configure_logging({"aspyx.di": logging.DEBUG})

from aspyx.di import DIException, injectable, order, on_init, on_running, on_destroy, inject_environment, inject, \
    Factory, create, module, Environment, PostProcessor, factory, requires_feature, conditional, requires_class

from di_import import ImportedModule, ImportedClass

### TEST

@injectable()
@order(10)
class SamplePostProcessor(PostProcessor):
    def process(self, instance: object, environment: Environment):
        pass #print(f"created a {instance}")

class Baa:
    def init(self):
        pass

class Foo:
    def __init__(self):
        self.inited = False

    @on_init()
    def init(self):
        self.inited = True

class Baz:
    def __init__(self):
        self.inited = False

    @on_init()
    def init(self):
        self.inited = True

@injectable()
class Bazong:
    def __init__(self):
        pass

class ConditionalBase:
    pass

@injectable()
@conditional(requires_feature("dev"))
class DevClass(ConditionalBase):
    def __init__(self):
        pass

@injectable()
@conditional(requires_class(DevClass))
class DevDependantClass:
    def __init__(self):
        pass

@injectable()
@conditional(requires_feature("prod"))
class ProdClass(ConditionalBase):
    def __init__(self):
        pass

@injectable()
@conditional(requires_class(ConditionalBase))
class RequiresBase:
    def __init__(self, base: ConditionalBase):
        pass

class Base:
    def __init__(self):
        pass

class Ambiguous:
    def __init__(self):
        pass

class Unknown:
    def __init__(self):
        pass#

@injectable(scope="request")
class NonSingleton:
    def __init__(self):
        super().__init__()

@injectable()
class Derived(Ambiguous):
    def __init__(self):
        super().__init__()

@injectable()
class Derived1(Ambiguous):
    def __init__(self):
        super().__init__()

@injectable()
class Bar(Base):
    def __init__(self, foo: Foo):
        super().__init__()

        self.bazong = None
        self.baz = None
        self.foo = foo
        self.inited = False
        self.running = False
        self.destroyed = False
        self.environment = None

    @create()
    def create_baa(self) -> Baa:
        return Baa()

    @on_init()
    def init(self):
        self.inited = True

    @on_running()
    def set_running(self):
        self.running = True

    @on_destroy()
    def destroy(self):
        self.destroyed = True

    @inject_environment()
    def init_environment(self, env: Environment):
        self.environment = env

    @inject()
    def set(self, baz: Baz, bazong: Bazong) -> None:
        self.baz = baz
        self.bazong = bazong

@factory()
class SampleFactory(Factory[Foo]):
    __slots__ = []

    def __init__(self):
        pass

    def create(self) -> Foo:
        return Foo()

@module()
class SimpleModule:
    # constructor

    def __init__(self):
        pass

    #TEST
    @create()
    def create_config(self) -> EnvConfigurationSource:
        return EnvConfigurationSource()

    # TES

    # create some beans

    @create()
    def create(self) -> Baz: #source: EnvConfigurationSource
        return Baz()

@module(imports=[SimpleModule, ImportedModule])
class ComplexModule:
    # constructor

    def __init__(self):
        pass

class TestDI(unittest.TestCase):
    testEnvironment = Environment(SimpleModule, features=["dev"])

    def test_thread_test(self):
        n_threads = 1
        iterations = 10000

        threads = []

        def worker(thread_id: int):
            env = Environment(SimpleModule, features=["dev"])

            for i in range(iterations):
                foo = env.get(Foo)

        for t_id in range(0, n_threads):
            thread = threading.Thread(target=worker, args=(t_id,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        print("All threads finished.")

    def test_conditional(self):
        env = TestDI.testEnvironment

        base = env.get(ConditionalBase)
        dev = env.get(DevClass)
        dep = env.get(DevDependantClass)

        try:
            env.get(ProdClass)
            self.fail("should not return conditional class")
        except Exception:
            pass

        self.assertIs(base, dev)
        self.assertIsNotNone(dev)
        self.assertIsNotNone(dep)

        # prod

        prod_environment = Environment(SimpleModule, features=["prod"])

        base = prod_environment.get(ConditionalBase)
        prod = prod_environment.get(ProdClass)

        self.assertIs(base, prod)
        self.assertIsNotNone(prod)

        print(prod_environment.report())

        # none

        try:
            no_feature_environment = Environment(SimpleModule)
            no_feature_environment = prod_environment.get(RequiresBase)
            self.fail("should not return conditional class")
        except Exception as e:
            print(e)
            pass

    def test_process_factory_instances(self):
        env = TestDI.testEnvironment

        print(env.report())
        print(env.parent.report())

        baz = env.get(Baz)
        foo = env.get(Foo)
        self.assertEqual(baz.inited, True)
        self.assertEqual(foo.inited, True)

    def test_inject_base_class(self):
        env = TestDI.testEnvironment

        base = env.get(Base)
        self.assertEqual(type(base), Bar)

    def test_inject_ambiguous_class(self):
        with self.assertRaises(DIException):
            env = TestDI.testEnvironment
            env.get(Ambiguous)

    def test_create_unknown(self):
        with self.assertRaises(DIException):
            env = TestDI.testEnvironment
            env.get(Unknown)

    def test_inject_constructor(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)
        baz = env.get(Baz)
        bazong = env.get(Bazong)
        foo = env.get(Foo)

        self.assertIsNotNone(bar)
        self.assertIs(bar.foo, foo)
        self.assertIs(bar.baz, baz)
        self.assertIs(bar.bazong, bazong)

    def test_factory(self):
        env = TestDI.testEnvironment
        foo = env.get(Foo)
        self.assertIsNotNone(foo)

    def test_create_factory(self):
        env = TestDI.testEnvironment
        baz = env.get(Baz)
        baa= env.get(Baa)
        self.assertIsNotNone(baz)
        self.assertIsNotNone(baa)

    def test_singleton(self):
        env = TestDI.testEnvironment

        # injectable

        bar = env.get(Bar)
        bar1 = env.get(Bar)
        self.assertIs(bar, bar1)

        # factory

        foo = env.get(Foo)
        foo1 = env.get(Foo)
        self.assertIs(foo,foo1)

        # create

        baz  = env.get(Baz)
        baz1 = env.get(Baz)
        self.assertIs(baz, baz1)

    def test_non_singleton(self):
        env = TestDI.testEnvironment

        ns = env.get(NonSingleton)
        ns1 = env.get(NonSingleton)

        self.assertIsNot(ns, ns1)

    def test_import_configurations(self):
        env = Environment(ImportedModule)

        imported = env.get(ImportedClass)
        self.assertIsNotNone(imported)

    def test_init(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        self.assertEqual(bar.inited, True)

    def test_running(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        self.assertEqual(bar.running, True)

    def test_destroy(self):
        env = TestDI.testEnvironment

        bar = env.get(Bar)

        env.destroy()

        self.assertEqual(bar.destroyed, True)

    def test_performance(self):
        env = TestDI.testEnvironment

        start = time.perf_counter()
        for _ in range(1000000):
            env.get(Bar)

        end = time.perf_counter()

        avg_ms = ((end - start) / 1000000) * 1000
        print(f"Average time per Bar creation: {avg_ms:.3f} ms")


if __name__ == '__main__':
    unittest.main()



"""
Test cases for the Configuration system in aspyx.di
"""
from __future__ import annotations

import unittest
from pathlib import Path

from aspyx.di.configuration import ConfigurationSource, ConfigurationManager, inject_value, EnvConfigurationSource, \
    YamlConfigurationSource

from aspyx.di import injectable, Environment, module, create


@module()
class SampleModule:
    # constructor

    def __init__(self):
        pass

    @create()
    def create_env_source(self) -> EnvConfigurationSource:
        return EnvConfigurationSource()

    @create()
    def create_yaml_source(self) -> YamlConfigurationSource:
        return YamlConfigurationSource(f"{Path(__file__).parent}/config.yaml"
)

@injectable()
class SampleConfigurationSource1(ConfigurationSource):
    # constructor

    def __init__(self):
        super().__init__()

    def load(self) -> dict:
        return {
            "a": 1, 
            "b": {}
            }

@injectable()
class SampleConfigurationSource2(ConfigurationSource):
    # constructor

    def __init__(self):
        super().__init__()

    def load(self) -> dict:
        return {
            "b": {
                "d": "2", 
                "e": 3, 
                "f": 4
                }
            }

@injectable(scope="request")
class Foo:
    def __init__(self):
        self.value = None
        self.value1 = None
        self.value2 = None

    @inject_value("b.d", 0)
    def set_value(self, value: int):
        self.value = value

    # will coerce
    @inject_value("b.e", 0)
    def set_value1(self, value: str):
        self.value1 = value

    @inject_value("b.z", "unknown")
    def set_value2(self, value: str):
        self.value2 = value

class TestConfiguration(unittest.TestCase):
    testEnvironment = Environment(SampleModule)

    def test_yaml(self):
        env = TestConfiguration.testEnvironment

        config = env.get(ConfigurationManager)

        port = config.get("server.port", int)
        self.assertEqual(port, 8000)

    def test_env(self):
        env = TestConfiguration.testEnvironment

        config = env.get(ConfigurationManager)

        os = config.get("HOME", str)
        self.assertIsNotNone(os)

    def test_configuration(self):
        env = TestConfiguration.testEnvironment

        config = env.get(ConfigurationManager)
        v1 = config.get("b.d", str)
        v2 = config.get("b.e", int, )
        v3 = config.get("b.z", str, "unknown")

        self.assertEqual(v1, "2")
        self.assertEqual(v2, 3)
        self.assertEqual(v3, "unknown")

    def test_configuration_injection(self):
        env = TestConfiguration.testEnvironment

        foo = env.get(Foo)

        self.assertEqual(foo.value, 2)

    def test_configuration_injection_coercion(self):
        env = TestConfiguration.testEnvironment

        foo = env.get(Foo)

        self.assertEqual(foo.value1, "3")

    def test_configuration_injection_default(self):
        env = TestConfiguration.testEnvironment

        foo = env.get(Foo)

        self.assertEqual(foo.value2, "unknown")


if __name__ == '__main__':
    unittest.main()



"""
This module provides tools for dynamic proxy creation and reflection
"""
from .proxy import DynamicProxy
from .reflection import Decorators, TypeDescriptor, DecoratorDescriptor

__all__ = [
    "DynamicProxy",
    "Decorators",
    "DecoratorDescriptor",
    "TypeDescriptor"
]



"""
YamlConfigurationSource - Loads variables from a YAML configuration file.
"""
import yaml

from .configuration import ConfigurationSource

class YamlConfigurationSource(ConfigurationSource):
    """
    YamlConfigurationSource loads variables from a YAML configuration file.
    """

    __slots__ = ["file"]

    # constructor

    def __init__(self, file: str):
        super().__init__()

        self.file = file

    # implement

    def load(self) -> dict:
        with open(self.file, "r") as file:
            return yaml.safe_load(file)



"""
Tests
"""
from __future__ import annotations

import unittest

from aspyx.di import injectable, module, Environment
from aspyx.di.di import DIException, inject


@injectable()
class Foo:
    def __init__(self, foo: Bar):
        pass

@injectable()
class Baz:
    def __init__(self):
        pass

@injectable()
class Bar:
    def __init__(self, foo: Foo):
        pass

    @inject()
    def set_baz(self, baz: Baz):
        pass


@module()
class SampleModule:
    # constructor

    def __init__(self):
        pass

class TestCycle(unittest.TestCase):
    def test_cycle(self):
        with self.assertRaises(DIException):
            Environment(SampleModule)


if __name__ == '__main__':
    unittest.main()



"""
Synchronized decorator and advice
"""
from __future__ import annotations

import threading
from weakref import WeakKeyDictionary

from aspyx.reflection import Decorators
from aspyx.di.aop import advice, around, methods, Invocation

def synchronized():
    """
    decorate methods to synchronize them based on an instance related `RLock`
    """
    def decorator(func):
        Decorators.add(func, synchronized)
        return func #

    return decorator

@advice
class SynchronizeAdvice:
    # constructor

    def __init__(self):
        self.locks = WeakKeyDictionary()

    # internal

    def get_lock(self, instance) -> threading.RLock:
        lock = self.locks.get(instance, None)
        if lock is None:
            lock = threading.RLock()
            self.locks[instance] = lock

        return lock

    # around

    @around(methods().decorated_with(synchronized))
    def synchronize_sync(self, invocation: Invocation):
        with self.get_lock(invocation.args[0]):
            return invocation.proceed()

    @around(methods().decorated_with(synchronized).that_are_async())
    async def synchronize_async(self, invocation: Invocation):
        with self.get_lock(invocation.args[0]):
            return await invocation.proceed_async()



"""
The dependency injection module provides a framework for managing dependencies and lifecycle of objects in Python applications.
"""
from __future__ import annotations

import inspect
import logging
import importlib
import pkgutil
import sys

from abc import abstractmethod, ABC
from enum import Enum
import threading
from typing import Type, Dict, TypeVar, Generic, Optional, cast, Callable, TypedDict

from aspyx.util import StringBuilder
from aspyx.reflection import Decorators, TypeDescriptor, DecoratorDescriptor

T = TypeVar("T")

class Factory(ABC, Generic[T]):
    """
    Abstract base class for factories that create instances of type T.
    """

    __slots__ = []

    @abstractmethod
    def create(self) -> T:
        pass

class DIException(Exception):
    """
    Exception raised for errors in the injector.
    """
    def __init__(self, message: str):
        super().__init__(message)

class DIRegistrationException(DIException):
    """
    Exception raised during the registration of dependencies.
    """
    def __init__(self, message: str):
        super().__init__(message)

class ProviderCollisionException(DIRegistrationException):
    def __init__(self, message: str, *providers: AbstractInstanceProvider):
        super().__init__(message)

        self.providers = providers

    def __str__(self):
        return f"[{self.args[0]} {self.providers[1].location()} collides with {self.providers[0].location()}"

class DIRuntimeException(DIException):
    """
    Exception raised during the runtime.
    """
    def __init__(self, message: str):
        super().__init__(message)

class AbstractInstanceProvider(ABC, Generic[T]):
    """
    An AbstractInstanceProvider is responsible to create instances.
    """
    @abstractmethod
    def get_module(self) -> str:
        """
        return the module name of the provider

        Returns:
            str: the module name of the provider
        """

    def get_host(self) -> Type[T]:
        """
        return the class which is responsible for creation ( e.g. the injectable class )

        Returns:
            Type[T]: the class which is responsible for creation
        """
        return type(self)

    @abstractmethod
    def get_type(self) -> Type[T]:
        """
        return the type of the created instance
        
        Returns:
            Type[T: the type]
        """

    @abstractmethod
    def is_eager(self) -> bool:
        """
        return True, if the provider will eagerly construct instances

        Returns:
            bool: eager flag
        """

    @abstractmethod
    def get_scope(self) -> str:
        """
        return the scope name

        Returns:
            str: the scope name
        """


    def get_dependencies(self) -> (list[Type],int):
        """
        return the types that i depend on ( for constructor or setter injection  ).
        The second tuple element is the number of parameters that a construction injection will require

        Returns:
            (list[Type],int): the type array and the number of parameters
        """
        return [],1

    @abstractmethod
    def create(self, environment: Environment, *args) -> T:
        """
        Create a new instance.

        Args:
            environment: the Environment
            *args: the required arguments

        Returns:
            T: the instance
        """

    def report(self) -> str:
        return str(self)

    def location(self) -> str:
        host = self.get_host()

        file = inspect.getfile(host)
        line = inspect.getsourcelines(host)[1]

        return f"{file}:{line}"

    def check_factories(self):
        pass


class InstanceProvider(AbstractInstanceProvider):
    """
    An InstanceProvider is able to create instances of type T.
    """
    __slots__ = [
        "host",
        "type",
        "eager",
        "scope"
    ]

    # constructor

    def __init__(self, host: Type, t: Type[T], eager: bool, scope: str):
        self.host = host
        self.type = t
        self.eager = eager
        self.scope = scope

    # implement AbstractInstanceProvider

    def get_host(self):
        return self.host

    def check_factories(self):
        pass

    def get_module(self) -> str:
        return self.host.__module__

    def get_type(self) -> Type[T]:
        return self.type

    def is_eager(self) -> bool:
        return self.eager

    def get_scope(self) -> str:
        return self.scope

    # public

    def module(self) -> str:
        return self.host.__module__

    @abstractmethod
    def create(self, environment: Environment, *args):
        pass

# we need this classes to bootstrap the system...
class SingletonScopeInstanceProvider(InstanceProvider):
    def __init__(self):
        super().__init__(SingletonScopeInstanceProvider, SingletonScope, False, "request")

    def create(self, environment: Environment, *args):
        return SingletonScope()

class EnvironmentScopeInstanceProvider(InstanceProvider):
    def __init__(self):
        super().__init__(SingletonScopeInstanceProvider, SingletonScope, False, "request")

    def create(self, environment: Environment, *args):
        return EnvironmentScope()

class RequestScopeInstanceProvider(InstanceProvider):
    def __init__(self):
        super().__init__(RequestScopeInstanceProvider, RequestScope, False, "singleton")

    def create(self, environment: Environment, *args):
        return RequestScope()

class AmbiguousProvider(AbstractInstanceProvider):
    """
    An AmbiguousProvider covers all cases, where fetching a class would lead to an ambiguity exception.
    """

    __slots__ = [
        "type",
        "providers",
    ]

    # constructor

    def __init__(self, type: Type, *providers: AbstractInstanceProvider):
        super().__init__()

        self.type = type
        self.providers = list(providers)

    # public

    def add_provider(self, provider: AbstractInstanceProvider):
        self.providers.append(provider)

    # implement

    def get_module(self) -> str:
        return self.type.__module__

    def get_type(self) -> Type[T]:
        return self.type

    def is_eager(self) -> bool:
        return False

    def get_scope(self) -> str:
        return "singleton"

    def create(self, environment: Environment, *args):
        raise DIException(f"multiple candidates for type {self.type}")

    def report(self) -> str:
        return "ambiguous: " + ",".join([provider.report() for provider in self.providers])

    def __str__(self):
        return f"AmbiguousProvider({self.type})"

class Scopes:
    # static data

    scopes : Dict[str, Type] = {}

    # class methods

    @classmethod
    def get(cls, scope: str, environment: Environment):
        scope_type = Scopes.scopes.get(scope, None)
        if scope_type is None:
            raise DIRegistrationException(f"unknown scope type {scope}")

        return environment.get(scope_type)

    @classmethod
    def register(cls, scope_type: Type, name: str):
        Scopes.scopes[name] = scope_type

class Scope:
    # properties

    __slots__ = [
    ]

    # constructor

    def __init__(self):
        pass

    # public

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        return provider.create(environment, *arg_provider())

class EnvironmentInstanceProvider(AbstractInstanceProvider):
    # properties

    __slots__ = [
        "environment",
        "scope_instance",
        "dependencies",
        "provider"
    ]

    # constructor

    def __init__(self, environment: Environment, provider: AbstractInstanceProvider):
        super().__init__()

        self.environment = environment
        self.provider = provider
        self.dependencies : list[AbstractInstanceProvider] = []
        self.scope_instance = Scopes.get(provider.get_scope(), environment)

    # public

    def print_tree(self, prefix=""):
        children = self.dependencies
        last_index = len(children) - 1
        print(prefix + "+- " + self.report())

        for i, child in enumerate(children):
            if i == last_index:
                # Last child
                child_prefix = prefix + "   "
            else:
                # Not last child
                child_prefix = prefix + "|  "


            cast(EnvironmentInstanceProvider, child).print_tree(child_prefix)

    # implement

    def resolve(self, context: Providers.ResolveContext):
        context.add(self)
        context.push(self)

        if not context.is_resolved(self):
            context.provider_dependencies[self] = [] #?

            type_and_params = self.provider.get_dependencies()
            params = type_and_params[1]
            for type in type_and_params[0]:
                if params > 0:
                    params -= 1
                    self.dependencies.append(context.get_provider(type)) # try/catch TODO

                provider = context.add_provider_dependency(self, type)
                if provider is not None:
                    provider.resolve(context)

        else:
            context.add(*context.get_provider_dependencies(self))

        context.pop()

    def get_module(self) -> str:
        return self.provider.get_module()

    def get_type(self) -> Type[T]:
        return self.provider.get_type()

    def is_eager(self) -> bool:
        return self.provider.is_eager()

    def get_scope(self) -> str:
        return self.provider.get_scope()

    def report(self) -> str:
        return self.provider.report()

    # own logic

    def create(self, environment: Environment, *args):
        return self.scope_instance.get(self.provider, self.environment, lambda: [provider.create(environment) for provider in self.dependencies]) # already scope property!

    def __str__(self):
        return f"EnvironmentInstanceProvider({self.provider})"

class ClassInstanceProvider(InstanceProvider):
    """
    A ClassInstanceProvider is able to create instances of type T by calling the class constructor.
    """

    __slots__ = [
        "params"
    ]

    # constructor

    def __init__(self, t: Type[T], eager: bool, scope = "singleton"):
        super().__init__(t, t, eager, scope)

        self.params = 0

    # implement

    def check_factories(self):
        register_factories(self.host)

    def get_dependencies(self) -> (list[Type],int):
        types : list[Type] = []

        # check constructor

        init = TypeDescriptor.for_type(self.type).get_method("__init__")
        if init is None:
            raise DIRegistrationException(f"{self.type.__name__} does not implement __init__")

        self.params = len(init.param_types)
        for param in init.param_types:
            types.append(param)

        # check @inject

        for method in TypeDescriptor.for_type(self.type).get_methods():
            if method.has_decorator(inject):
                for param in method.param_types:
                    if not Providers.is_registered(param):
                        raise DIRegistrationException(f"{self.type.__name__}.{method.method.__name__} declares an unknown parameter type {param.__name__}")

                    types.append(param)
        # done

        return types, self.params

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        return environment.created(self.type(*args[:self.params]))

    def report(self) -> str:
        return f"{self.host.__name__}.__init__"

    # object

    def __str__(self):
        return f"ClassInstanceProvider({self.type.__name__})"

class FunctionInstanceProvider(InstanceProvider):
    """
    A FunctionInstanceProvider is able to create instances of type T by calling specific methods annotated with 'create".
    """

    __slots__ = [
        "method"
    ]

    # constructor

    def __init__(self, clazz : Type, method: TypeDescriptor.MethodDescriptor, eager = True, scope = "singleton"):
        super().__init__(clazz, method.return_type, eager, scope)

        self.method : TypeDescriptor.MethodDescriptor = method

    # implement

    def get_dependencies(self) -> (list[Type],int):
        return [self.host, *self.method.param_types], 1 + len(self.method.param_types)

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        instance = self.method.method(*args) # args[0]=self

        return environment.created(instance)

    def report(self) -> str:
        return f"{self.host.__name__}.{self.method.get_name()}({', '.join(t.__name__ for t in self.method.param_types)}) -> {self.type.__qualname__}"

    def __str__(self):
        return f"FunctionInstanceProvider({self.host.__name__}.{self.method.get_name()}({', '.join(t.__name__ for t in self.method.param_types)}) -> {self.type.__name__})"

class FactoryInstanceProvider(InstanceProvider):
    """
    A FactoryInstanceProvider is able to create instances of type T by calling registered Factory instances.
    """

    __slots__ = []

    # class method

    @classmethod
    def get_factory_type(cls, clazz):
        return TypeDescriptor.for_type(clazz).get_method("create", local=True).return_type

    # constructor

    def __init__(self, factory: Type, eager: bool, scope: str):
        super().__init__(factory, FactoryInstanceProvider.get_factory_type(factory), eager, scope)

    # implement

    def get_dependencies(self) -> (list[Type],int):
        return [self.host],1

    def create(self, environment: Environment, *args):
        Environment.logger.debug("%s create class %s", self, self.type.__qualname__)

        return environment.created(args[0].create())

    def report(self) -> str:
        return f"{self.host.__name__}.create() -> {self.type.__name__} "

    def __str__(self):
        return f"FactoryInstanceProvider({self.host.__name__} -> {self.type.__name__})"


class Lifecycle(Enum):
    """
    This enum defines the lifecycle phases that can be processed by lifecycle processors.
    Phases are:

    - ON_INJECT
    - ON_INIT
    - ON_RUNNING
    - ON_DESTROY
    """

    __slots__ = []

    ON_INJECT  = 0
    ON_INIT    = 1
    ON_RUNNING = 2
    ON_DESTROY = 3

class LifecycleProcessor(ABC):
    """
    A LifecycleProcessor is used to perform any side effects on managed objects during different lifecycle phases.
    """
    __slots__ = [
        "order"
    ]

    # constructor

    def __init__(self):
        self.order = 0
        if TypeDescriptor.for_type(type(self)).has_decorator(order):
            self.order =  TypeDescriptor.for_type(type(self)).get_decorator(order).args[0]

    # methods

    @abstractmethod
    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        pass

class PostProcessor(LifecycleProcessor):
    """
    Base class for custom post processors that are executed after object creation.
    """
    __slots__ = []


    def process(self, instance: object, environment: Environment):
        pass

    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        if lifecycle == Lifecycle.ON_INIT:
            self.process(instance, environment)


class Providers:
    """
    The Providers class is a static class that manages the registration and resolution of InstanceProviders.
    """
    # local class

    class ResolveContext:
        __slots__ = [
            "dependencies",
            "providers",
            "provider_dependencies",
            "path"
        ]

        # constructor

        def __init__(self, providers: Dict[Type, EnvironmentInstanceProvider]):
            self.dependencies : list[EnvironmentInstanceProvider] = []
            self.providers = providers
            self.path = []
            self.provider_dependencies : dict[EnvironmentInstanceProvider, list[EnvironmentInstanceProvider]] = {}

        # public

        def is_resolved(self, provider: EnvironmentInstanceProvider) -> bool:
            return self.provider_dependencies.get(provider, None) is not None

        def get_provider_dependencies(self, provider: EnvironmentInstanceProvider) -> list[EnvironmentInstanceProvider]:
            return self.provider_dependencies[provider]

        def add_provider_dependency(self, provider: EnvironmentInstanceProvider, type: Type) -> Optional[EnvironmentInstanceProvider]:
            provider_dependencies = self.provider_dependencies.get(provider, None)
            if provider_dependencies is None:
                provider_dependencies = []
                self.provider_dependencies[provider] = provider_dependencies

            provider = self.get_provider(type)

            if any(issubclass(provider.get_type(), dependency.get_type()) for dependency in provider_dependencies):
                return None

            provider_dependencies.append(provider)

            return provider

        def push(self, provider):
            self.path.append(provider)

        def pop(self):
            self.path.pop()

        def next(self):
            self.path.clear()
            self.dependencies.clear()

        def get_provider(self, type: Type) -> EnvironmentInstanceProvider:
            provider = self.providers.get(type, None)
            if provider is None:
                raise DIRegistrationException(f"Provider for {type} is not defined")

            return provider

        def add(self, *providers: EnvironmentInstanceProvider):
            for provider in providers:
                if next((p for p in self.dependencies if p.get_type() is provider.get_type()), None) is not None:
                    raise DIRegistrationException(self.cycle_report(provider))

                self.dependencies.append(provider)

        def cycle_report(self, provider: AbstractInstanceProvider):
            cycle = ""

            first = True
            for p in self.path:
                if not first:
                    cycle += " -> "

                first = False

                cycle += f"{p.report()}"

            cycle += f" <> {provider.report()}"

            return cycle


    # class properties

    check : list[AbstractInstanceProvider] = []
    providers : Dict[Type,list[AbstractInstanceProvider]] = {}

    resolved = False

    @classmethod
    def register(cls, provider: AbstractInstanceProvider):
        Environment.logger.debug("register provider %s(%s)", provider.get_type().__qualname__, provider.get_type().__name__)

        Providers.check.append(provider)
        candidates = Providers.providers.get(provider.get_type(), None)
        if candidates is None:
            Providers.providers[provider.get_type()] = [provider]
        else:
            candidates.append(provider)

    @classmethod
    def is_registered(cls,type: Type) -> bool:
        return Providers.providers.get(type, None) is not None

    # add factories lazily

    @classmethod
    def check_factories(cls):
        for check in Providers.check:
            check.check_factories()

        Providers.check.clear()

    @classmethod
    def filter(cls, environment: Environment, provider_filter: Callable) -> Dict[Type,AbstractInstanceProvider]:
        cache: Dict[Type,AbstractInstanceProvider] = {}

        context: ConditionContext = {
            "requires_feature": environment.has_feature,
            "requires_class": lambda clazz : cache.get(clazz, None) is not None # ? only works if the class is in the cache already?
        }

        Providers.check_factories() # check for additional factories

        # local methods

        def filter_type(clazz: Type) -> Optional[AbstractInstanceProvider]:
            result = None
            for provider in Providers.providers[clazz]:
                if provider_applies(provider):
                    if result is not None:
                        raise ProviderCollisionException(f"type {clazz.__name__} already registered", result, provider)

                    result = provider

            return result

        def provider_applies(provider: AbstractInstanceProvider) -> bool:
            # is it in the right module?

            if not provider_filter(provider):
                return False

            # check conditionals

            descriptor = TypeDescriptor.for_type(provider.get_host())
            if descriptor.has_decorator(conditional):
                conditions: list[Condition] = [*descriptor.get_decorator(conditional).args]
                for condition in conditions:
                    if not condition.apply(context):
                        return False

                return True

            return True

        def is_injectable(type: Type) -> bool:
            if type is object:
                return False

            if inspect.isabstract(type):
                return False

            return True

        def cache_provider_for_type(provider: AbstractInstanceProvider, type: Type):
            existing_provider = cache.get(type)
            if existing_provider is None:
                cache[type] = provider

            else:
                if type is provider.get_type():
                    raise ProviderCollisionException(f"type {type.__name__} already registered", existing_provider, provider)

                if existing_provider.get_type() is not type:
                    # only overwrite if the existing provider is not specific

                    if isinstance(existing_provider, AmbiguousProvider):
                        cast(AmbiguousProvider, existing_provider).add_provider(provider)
                    else:
                        cache[type] = AmbiguousProvider(type, existing_provider, provider)

            # recursion

            for super_class in type.__bases__:
                if is_injectable(super_class):
                    cache_provider_for_type(provider, super_class)

        # filter conditional providers and fill base classes as well

        for provider_type, _ in Providers.providers.items():
            matching_provider = filter_type(provider_type)
            if matching_provider is not None:
                cache_provider_for_type(matching_provider, provider_type)

        # replace by EnvironmentInstanceProvider

        mapped = {}
        result = {}
        for provider_type, provider in cache.items():
            environment_provider = mapped.get(provider, None)
            if environment_provider is None:
                environment_provider = EnvironmentInstanceProvider(environment,  provider)
                mapped[provider] = environment_provider

            result[provider_type] = environment_provider

        # and resolve

        providers = result
        if environment.parent is not None:
            providers = providers | environment.parent.providers # add parent providers

        provider_context = Providers.ResolveContext(providers)
        for provider in mapped.values():
            provider.resolve(provider_context)
            provider_context.next() # clear dependencies

        # done

        return result

def register_factories(cls: Type):
    descriptor = TypeDescriptor.for_type(cls)

    for method in descriptor.get_methods():
        if method.has_decorator(create):
            create_decorator = method.get_decorator(create)
            return_type = method.return_type
            if return_type is None:
                raise DIRegistrationException(f"{cls.__name__}.{method.method.__name__} expected to have a return type")

            Providers.register(FunctionInstanceProvider(cls, method, create_decorator.args[0], create_decorator.args[1]))
def order(prio = 0):
    def decorator(cls):
        Decorators.add(cls, order, prio)

        return cls

    return decorator

def injectable(eager=True, scope="singleton"):
    """
    Instances of classes that are annotated with @injectable can be created by an Environment.
    """
    def decorator(cls):
        Decorators.add(cls, injectable)

        Providers.register(ClassInstanceProvider(cls, eager, scope))

        return cls

    return decorator

def factory(eager=True, scope="singleton"):
    """
    Decorator that needs to be used on a class that implements the Factory interface.

    Args:
        eager (bool): If True, the corresponding object will be created eagerly when the environment is created.
        scope (str): The scope of the factory, e.g. "singleton", "request", "environment".
    """
    def decorator(cls):
        Decorators.add(cls, factory)

        Providers.register(ClassInstanceProvider(cls, eager, scope))
        Providers.register(FactoryInstanceProvider(cls, eager, scope))

        return cls

    return decorator

def create(eager=True, scope="singleton"):
    """
    Any method annotated with @create will be registered as a factory method.

    Args:
        eager (bool): If True, the corresponding object will be created eagerly when the environment is created.
        scope (str): The scope of the factory, e.g. "singleton", "request", "environment".
    """
    def decorator(func):
        Decorators.add(func, create, eager, scope)
        return func

    return decorator

def on_init():
    """
    Methods annotated with `@on_init` will be called when the instance is created."""
    def decorator(func):
        Decorators.add(func, on_init)
        return func

    return decorator

def on_running():
    """
    Methods annotated with `@on_running` will be called when the container up and running."""
    def decorator(func):
        Decorators.add(func, on_running)
        return func

    return decorator

def on_destroy():
    """
    Methods annotated with `@on_destroy` will be called when the instance is destroyed.
    """
    def decorator(func):
        Decorators.add(func, on_destroy)
        return func

    return decorator

def module(imports: Optional[list[Type]] = None):
    """
    This annotation is used to mark classes that control the discovery process of injectables based on their location
    relative to the module of the class. All `@injectable`s and `@factory`s that are located in the same or any sub-module will
    be registered and managed accordingly.

    Args:
        imports (Optional[list[Type]]): Optional list of imported module types
    """
    def decorator(cls):
        Providers.register(ClassInstanceProvider(cls, True))

        Decorators.add(cls, module, imports)
        Decorators.add(cls, injectable) # do we need that?

        return cls

    return decorator

def inject():
    """
    Methods annotated with @inject will be called with the required dependencies injected.
    """
    def decorator(func):
        Decorators.add(func, inject)
        return func

    return decorator

def inject_environment():
    """
    Methods annotated with @inject_environment will be called with the Environment instance injected.
    """
    def decorator(func):
        Decorators.add(func, inject_environment)
        return func

    return decorator

# conditional stuff

class ConditionContext(TypedDict):
    requires_feature: Callable[[str], bool]
    requires_class: Callable[[Type], bool]

class Condition(ABC):
    @abstractmethod
    def apply(self, context: ConditionContext) -> bool:
        pass

class FeatureCondition(Condition):
    def __init__(self, feature: str):
        super().__init__()

        self.feature = feature

    def apply(self, context: ConditionContext) -> bool:
        return context["requires_feature"](self.feature)

class ClassCondition(Condition):
    def __init__(self, clazz: Type):
        super().__init__()

        self.clazz = clazz

    def apply(self, context: ConditionContext) -> bool:
        return context["requires_class"](self.clazz)

def requires_feature(feature: str):
    return FeatureCondition(feature)

def requires_class(clazz: Type):
    return ClassCondition(clazz)

def conditional(*conditions: Condition):
    def decorator(cls):
        Decorators.add(cls, conditional, *conditions)

        return cls

    return decorator

class Environment:
    """
    Central class that manages the lifecycle of instances and their dependencies.

    Usage:

    ```python
    @injectable()
    class Foo:
        def __init__(self):

    @module()
    class Module:
        def __init__(self):
            pass

    environment = Environment(Module)

    foo = environment.get(Foo)  # will create an instance of Foo
    ```
    """

    # static data

    logger = logging.getLogger("aspyx.di")  # __name__ = module name

    instance : 'Environment' = None

    __slots__ = [
        "type",
        "providers",
        "lifecycle_processors",
        "parent",
        "features",
        "instances"
    ]

    # constructor

    def __init__(self, env: Type, features: list[str] = [], parent : Optional[Environment] = None):
        """
        Creates a new Environment instance.

        Args:
            env (Type): The environment class that controls the scanning of managed objects.
            parent (Optional[Environment]): Optional parent environment, whose objects are inherited.
        """

        Environment.logger.debug("create environment for class %s", env.__qualname__)

        # initialize

        self.type = env
        self.parent = parent
        if self.parent is None and env is not Boot:
            self.parent = Boot.get_environment() # inherit environment including its manged instances!

        self.features = features
        self.providers: Dict[Type, AbstractInstanceProvider] = {}
        self.instances = []
        self.lifecycle_processors: list[LifecycleProcessor] = []

        if self.parent is not None:
            # inherit providers from parent

            for provider_type, inherited_provider in self.parent.providers.items():
                if inherited_provider.get_scope() == "environment":
                    # replace with own environment instance provider
                    self.providers[provider_type] = EnvironmentInstanceProvider(self, cast(EnvironmentInstanceProvider, inherited_provider).provider)
                else:
                    self.providers[provider_type] = inherited_provider

            # inherit processors as is unless they have an environment scope

            for processor in self.parent.lifecycle_processors:
                if self.providers[type(processor)].get_scope() != "environment":
                    self.lifecycle_processors.append(processor)
                else:
                    # create and remember
                    self.lifecycle_processors.append(self.get(type(processor)))
        else:
            self.providers[SingletonScope] = SingletonScopeInstanceProvider()
            self.providers[RequestScope]   = RequestScopeInstanceProvider()
            self.providers[EnvironmentScope] = EnvironmentScopeInstanceProvider()

        Environment.instance = self

        prefix_list : list[str] = []

        loaded = set()

        def add_provider(type: Type, provider: AbstractInstanceProvider):
            Environment.logger.debug("\tadd provider %s for %s", provider, type)

            self.providers[type] = provider

        def get_type_package(type: Type):
            module_name = type.__module__
            module = sys.modules.get(module_name)

            if not module:
                raise ImportError(f"Module {module_name} not found")

            # Try to get the package

            package = getattr(module, '__package__', None)

            # Fallback: if module is __main__, try to infer from the module name if possible

            if not package:
                if module_name == '__main__':
                    # Try to resolve real name via __file__
                    path = getattr(module, '__file__', None)
                    if path:
                        Environment.logger.warning(
                            "Module is __main__; consider running via -m to preserve package context")
                    return ''

                # Try to infer package name from module name

                parts = module_name.split('.')
                if len(parts) > 1:
                    return '.'.join(parts[:-1])

            return package or ''

        def import_package(name: str):
            """Import a package and all its submodules recursively."""
            package = importlib.import_module(name)
            results = {name: package}

            if hasattr(package, '__path__'):  # it's a package, not a single file
                for finder, name, ispkg in pkgutil.walk_packages(package.__path__, prefix=package.__name__ + "."):
                    try:
                        loaded = sys.modules

                        if loaded.get(name, None) is None:
                            Environment.logger.debug("import module %s", name)

                            submodule = importlib.import_module(name)
                            results[name] = submodule
                        else:
                            # skip import
                            results[name] = loaded[name]

                    except Exception as e:
                        Environment.logger.info("failed to import module %s due to %s", name, str(e))

            return results

        def load_environment(env: Type):
            if env not in loaded:
                Environment.logger.debug("load environment %s", env.__qualname__)

                loaded.add(env)

                # sanity check

                decorator = TypeDescriptor.for_type(env).get_decorator(module)
                if decorator is None:
                    raise DIRegistrationException(f"{env.__name__} is not an environment class")

                # package

                package_name = get_type_package(env)

                # recursion

                for import_environment in decorator.args[0] or []:
                    load_environment(import_environment)

                # import package

                if package_name is not None and len(package_name) > 0: # files outside of a package return None pr ""
                    import_package(package_name)

                # filter and load providers according to their module

                module_prefix = package_name
                if len(module_prefix) == 0:
                    module_prefix = env.__module__

                prefix_list.append(module_prefix)

        # go

        load_environment(env)

        # filter according to the prefix list

        def filter_provider(provider: AbstractInstanceProvider) -> bool:
            for prefix in prefix_list:
                if provider.get_host().__module__.startswith(prefix):
                    return True

            return False

        self.providers.update(Providers.filter(self, filter_provider))

        # construct eager objects for local providers

        for provider in set(self.providers.values()):
            if isinstance(provider, EnvironmentInstanceProvider):
                provider.print_tree()

            if provider.is_eager():
                provider.create(self)

        # running callback

        for instance in self.instances:
            self.execute_processors(Lifecycle.ON_RUNNING, instance)

    def is_registered_type(self, type: Type) -> bool:
        provider = self.providers.get(type, None)
        return provider is not None and not isinstance(provider, AmbiguousProvider)

    def registered_types(self,  predicate: Callable[[Type], bool]) -> list[Type]:
        return [provider.get_type() for provider in self.providers.values()
                if predicate(provider.get_type())]

    # internal

    def has_feature(self, feature: str) -> bool:
        return feature in self.features

    def execute_processors(self, lifecycle: Lifecycle, instance: T) -> T:
        for processor in self.lifecycle_processors:
            processor.process_lifecycle(lifecycle, instance, self)

        return instance

    def created(self, instance: T) -> T:
        # remember lifecycle processors

        if isinstance(instance, LifecycleProcessor):
            self.lifecycle_processors.append(instance)

            # sort immediately

            self.lifecycle_processors.sort(key=lambda processor: processor.order)

        # remember instance

        self.instances.append(instance)

        # execute processors

        self.execute_processors(Lifecycle.ON_INJECT, instance)
        self.execute_processors(Lifecycle.ON_INIT, instance)

        return instance

    # public

    def report(self):
        builder = StringBuilder()

        builder.append(f"Environment {self.type.__name__}")

        if self.parent is not None:
            builder.append(f" parent {self.parent.type.__name__}")

        builder.append("\n")

        # post processors

        builder.append("Processors \n")
        for processor in self.lifecycle_processors:
            builder.append(f"- {processor.__class__.__name__}\n")

        # providers

        builder.append("Providers \n")
        for result_type, provider in self.providers.items():
            if isinstance(provider, EnvironmentInstanceProvider):
                if cast(EnvironmentInstanceProvider, provider).environment is self:
                    builder.append(f"- {result_type.__name__}: {provider.report()}\n")

        # instances

        builder.append("Instances \n")

        result = {}
        for obj in self.instances:
            cls = type(obj)
            result[cls] = result.get(cls, 0) + 1

        for cls, count in result.items():
            builder.append(f"- {cls.__name__}: {count} \n")

        # done

        result = str(builder)

        return result

    def destroy(self):
        """
        destroy all managed instances by calling the appropriate lifecycle methods
        """
        for instance in self.instances:
            self.execute_processors(Lifecycle.ON_DESTROY, instance)

        self.instances.clear() # make the cy happy

    def get(self, type: Type[T]) -> T:
        """
        Create or return a cached instance for the given type.

        Args:
            type (Type): The desired type

        Returns:
            T: The requested instance
        """
        provider = self.providers.get(type, None)
        if provider is None:
            Environment.logger.error("%s is not supported", type)
            raise DIRuntimeException(f"{type} is not supported")

        return provider.create(self)

    def __str__(self):
        return f"Environment({self.type.__name__})"

class LifecycleCallable:
    __slots__ = [
        "decorator",
        "lifecycle",
        "order"
    ]

    def __init__(self, decorator, lifecycle: Lifecycle):
        self.decorator = decorator
        self.lifecycle = lifecycle
        self.order = 0

        if TypeDescriptor.for_type(type(self)).has_decorator(order):
            self.order = TypeDescriptor.for_type(type(self)).get_decorator(order).args[0]

        AbstractCallableProcessor.register(self)

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return []


class AbstractCallableProcessor(LifecycleProcessor):
    # local classes

    class MethodCall:
        __slots__ = [
            "decorator",
            "method",
            "lifecycle_callable"
        ]

        # constructor

        def __init__(self, method: TypeDescriptor.MethodDescriptor, decorator: DecoratorDescriptor, lifecycle_callable: LifecycleCallable):
            self.decorator = decorator
            self.method = method
            self.lifecycle_callable = lifecycle_callable

        def execute(self, instance, environment: Environment):
            self.method.method(instance, *self.lifecycle_callable.args(self.decorator, self.method, environment))

        def __str__(self):
            return f"MethodCall({self.method.method.__name__})"

    # static data

    lock = threading.RLock()
    callables : Dict[object, LifecycleCallable] = {}
    cache : Dict[Type, list[list[AbstractCallableProcessor.MethodCall]]] = {}

    # static methods

    @classmethod
    def register(cls, callable: LifecycleCallable):
        AbstractCallableProcessor.callables[callable.decorator] = callable

    @classmethod
    def compute_callables(cls, type: Type) -> list[list[AbstractCallableProcessor.MethodCall]]:
        descriptor = TypeDescriptor.for_type(type)

        result = [[], [], [], []]  # per lifecycle

        for method in descriptor.get_methods():
            for decorator in method.decorators:
                callable = AbstractCallableProcessor.callables.get(decorator.decorator)
                if callable is not None:  # any callable for this decorator?
                    result[callable.lifecycle.value].append(
                        AbstractCallableProcessor.MethodCall(method, decorator, callable))

        # sort according to order

        result[0].sort(key=lambda call: call.lifecycle_callable.order)
        result[1].sort(key=lambda call: call.lifecycle_callable.order)
        result[2].sort(key=lambda call: call.lifecycle_callable.order)
        result[3].sort(key=lambda call: call.lifecycle_callable.order)

        # done

        return result

    @classmethod
    def callables_for(cls, type: Type) -> list[list[AbstractCallableProcessor.MethodCall]]:
        callables = AbstractCallableProcessor.cache.get(type, None)
        if callables is None:
            with AbstractCallableProcessor.lock:
                callables = AbstractCallableProcessor.cache.get(type, None)
                if callables is None:
                    callables = AbstractCallableProcessor.compute_callables(type)
                    AbstractCallableProcessor.cache[type] = callables

        return callables

    # constructor

    def __init__(self, lifecycle: Lifecycle):
        super().__init__()

        self.lifecycle = lifecycle

    # implement

    def process_lifecycle(self, lifecycle: Lifecycle, instance: object, environment: Environment) -> object:
        if lifecycle is self.lifecycle:
            callables = self.callables_for(type(instance))

            for callable in callables[lifecycle.value]:
                callable.execute(instance, environment)

@injectable()
@order(1)
class OnInjectCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_INJECT)

@injectable()
@order(2)
class OnInitCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_INIT)

@injectable()
@order(3)
class OnRunningCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_RUNNING)

@injectable()
@order(4)
class OnDestroyCallableProcessor(AbstractCallableProcessor):
    def __init__(self):
        super().__init__(Lifecycle.ON_DESTROY)

# the callables

@injectable()
@order(1000)
class OnInitLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_init, Lifecycle.ON_INIT)

@injectable()
@order(1001)
class OnDestroyLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_destroy, Lifecycle.ON_DESTROY)

@injectable()
@order(1002)
class OnRunningLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(on_running, Lifecycle.ON_RUNNING)

@injectable()
@order(9)
class EnvironmentAwareLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(inject_environment, Lifecycle.ON_INJECT)

    def args(self, decorator: DecoratorDescriptor, method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [environment]

@injectable()
@order(10)
class InjectLifecycleCallable(LifecycleCallable):
    __slots__ = []

    def __init__(self):
        super().__init__(inject, Lifecycle.ON_INJECT)

    # override

    def args(self, decorator: DecoratorDescriptor,  method: TypeDescriptor.MethodDescriptor, environment: Environment):
        return [environment.get(type) for type in method.param_types]

def scope(name: str, register=True):
    def decorator(cls):
        Scopes.register(cls, name)

        Decorators.add(cls, scope)

        if register:
            Providers.register(ClassInstanceProvider(cls, eager=True, scope="request"))

        return cls

    return decorator

@scope("request", register=False)
class RequestScope(Scope):
    # properties

    __slots__ = [
    ]

    # public

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        return provider.create(environment, *arg_provider())

@scope("singleton", register=False)
class SingletonScope(Scope):
    # properties

    __slots__ = [
        "value",
        "lock"
    ]

    # constructor

    def __init__(self):
        super().__init__()

        self.value = None
        self.lock = threading.RLock()

    # override

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[],list]):
        if self.value is None:
            with self.lock:
                if self.value is None:
                    self.value = provider.create(environment, *arg_provider())

        return self.value

@scope("environment", register=False)
class EnvironmentScope(SingletonScope):
    # properties

    __slots__ = [
    ]

    # constructor

    def __init__(self):
        super().__init__()


@scope("thread")
class ThreadScope(Scope):
    __slots__ = [
        "_local"
    ]

    # constructor

    def __init__(self):
        super().__init__()
        self._local = threading.local()

    def get(self, provider: AbstractInstanceProvider, environment: Environment, arg_provider: Callable[[], list]):
        if not hasattr(self._local, "value"):
            self._local.value = provider.create(environment, *arg_provider())

        return self._local.value

# internal class that is required to import technical instance providers

@module()
class Boot:
    # class

    environment = None

    @classmethod
    def get_environment(cls):
        if Boot.environment is None:
            Boot.environment = Environment(Boot)

        return Boot.environment

    # properties

    __slots__ = []

    # constructor

    def __init__(self):
        pass



"""
EnvConfigurationSource - Loads environment variables as configuration source.
"""
import os

from dotenv import load_dotenv

from .configuration import ConfigurationSource

class EnvConfigurationSource(ConfigurationSource):
    """
    EnvConfigurationSource loads all environment variables.
    """

    __slots__ = []

    # constructor

    def __init__(self):
        super().__init__()

    # implement

    def load(self) -> dict:
        def merge_dicts(a, b):
            """Recursively merges b into a"""
            for key, value in b.items():
                if isinstance(value, dict) and key in a and isinstance(a[key], dict):
                    merge_dicts(a[key], value)
                else:
                    a[key] = value
            return a

        def explode_key(key, value):
            """Explodes keys with '.' or '/' into nested dictionaries"""
            parts = key.replace('/', '.').split('.')
            d = current = {}
            for part in parts[:-1]:
                current[part] = {}
                current = current[part]
            current[parts[-1]] = value
            return d

        exploded = {}

        load_dotenv()

        for key, value in os.environ.items():
            if '.' in key or '/' in key:
                partial = explode_key(key, value)
                merge_dicts(exploded, partial)
            else:
                exploded[key] = value

        return exploded



"""
Utility class for Java lovers
"""
class StringBuilder:
    """
    A StringBuilder is used to build a string by multiple append calls.
    """
    ___slots__ = ("_parts",)

    # constructor

    def __init__(self):
        self._parts = []

    # public

    def append(self, s: str) -> "StringBuilder":
        """
        append a string to the end of the string builder

        Args:
            s (str): the string

        Returns:
            StringBuilder: self
        """
        self._parts.append(str(s))

        return self

    def extend(self, iterable) -> "StringBuilder":
        for s in iterable:
            self._parts.append(str(s))

        return self

    def clear(self):
        """
        clear the content
        """
        self._parts.clear()

    # object

    def __str__(self):
        return ''.join(self._parts)



"""
Exception handling code
"""
from threading import RLock
from typing import Any, Callable, Dict, Optional, Type

from aspyx.di import injectable, Environment, inject_environment, on_running
from aspyx.reflection import Decorators, TypeDescriptor
from aspyx.threading import ThreadLocal


def exception_handler():
    """
    This annotation is used to mark classes that container handlers for exceptions
    """
    def decorator(cls):
        Decorators.add(cls, exception_handler)

        ExceptionManager.exception_handler_classes.append(cls)

        return cls

    return decorator

def handle():
    """
    Any method annotated with @handle will be registered as an exception handler method.
    """
    def decorator(func):
        Decorators.add(func, handle)
        return func

    return decorator

class Handler:
    # constructor

    def __init__(self, type_: Type, instance: Any, handler: Callable):
        self.type = type_
        self.instance = instance
        self.handler = handler

    def handle(self, exception: BaseException) -> BaseException:
        result = self.handler(self.instance, exception)

        if result is not None:
            return result
        else:
            return exception

class Chain:
    # constructor

    def __init__(self, handler: Handler, next: Optional[Handler] = None):
        self.handler = handler
        self.next = next

    # public

    def handle(self, exception: BaseException) -> BaseException:
        return self.handler.handle(exception)

class Invocation:
    def __init__(self, exception: BaseException, chain: Chain):
        self.exception = exception
        self.chain = chain
        self.current = self.chain

@injectable()
class ExceptionManager:
    """
    An exception manager collects all registered handlers and is able to handle an exception
    by dispatching it to the most applicable handler ( according to mro )
    """
    # static data

    exception_handler_classes = []

    invocation = ThreadLocal[Invocation]()

    # class methods

    @classmethod
    def proceed(cls) -> BaseException:
        """
        proceed with the next most applicable handler

        Returns:
            BaseException: the resulting exception

        """
        invocation = cls.invocation.get()

        invocation.current = invocation.current.next
        if invocation.current is not None:
            return invocation.current.handle(invocation.exception)
        else:
            return invocation.exception

    # constructor

    def __init__(self):
        self.environment : Optional[Environment] = None
        self.handler : list[Handler] = []
        self.cache: Dict[Type, Chain] = {}
        self.lock = RLock()

    # internal

    @inject_environment()
    def set_environment(self, environment: Environment):
        self.environment = environment

    @on_running()
    def setup(self):
        for handler_class in self.exception_handler_classes:
            type_descriptor = TypeDescriptor.for_type(handler_class)
            instance = self.environment.get(handler_class)

            # analyze methods

            for method in type_descriptor.get_methods():
                if method.has_decorator(handle):
                    if len(method.param_types) == 1:
                        exception_type = method.param_types[0]

                        self.handler.append(Handler(
                            exception_type,
                            instance,
                            method.method,
                        ))
                    else:
                        print(f"handler {method.method} expected to have one parameter")

    def get_handlers(self, clazz: Type) -> Optional[Chain]:
        chain = self.cache.get(clazz, None)
        if chain is None:
            with self.lock:
                chain = self.cache.get(clazz, None)
                if chain is None:
                    chain = self.compute_handlers(clazz)
                    self.cache[clazz] = chain

        return chain

    def compute_handlers(self, clazz: Type) -> Optional[Chain]:
        mro = clazz.mro()

        chain = []

        for type in mro:
            handler = next((handler for handler in self.handler if handler.type is type), None)
            if handler:
                chain.append(Chain(handler))

        # chain

        for i in range(0, len(chain) - 1):
            chain[i].next = chain[i + 1]

        if len(chain) > 0:
            return chain[0]
        else:
            return None

    def handle(self, exception: BaseException) -> BaseException:
        """
        handle an exception by invoking the most applicable handler (according to mro)
        and return a possible modified exception as a result.

        Args:
            exception (BaseException): the exception

        Returns:
            BaseException: the resulting exception
        """
        chain = self.get_handlers(type(exception))
        if chain is not None:
            self.invocation.set(Invocation(exception, chain))
            try:
                return chain.handle(exception)
            finally:
                self.invocation.clear()
        else:
            return exception # hmmm?



"""
The AOP module gives you the possibility to define aspects that will participate in method execution flows.

**Example**: all method executions of methods named "foo" will include a `before` aspect, that will be executed before the original method

```python
@advice
class Advice:
   @before(methods().named("foo"))
   def before_call(self, invocation: Invocation):
      ...

```

Note, that this requires that both the advice and the targeted methods need to be managed by an environment.
"""
from .aop import before, after, classes, around, error, advice, methods, Invocation, AspectTarget
__all__ = [
    "before",
    "after",
    "around",
    "error",
    "advice",
    "classes",
    "methods",
    "Invocation",
    "AspectTarget"
]



"""
This module contains functionality to read configuration values from different sources and to retrieve or inject them.
"""
from .configuration import ConfigurationManager, ConfigurationSource, inject_value
from .env_configuration_source import EnvConfigurationSource
from .yaml_configuration_source import YamlConfigurationSource

__all__ = [
    "ConfigurationManager",
    "ConfigurationSource",
    "EnvConfigurationSource",
    "YamlConfigurationSource",
    "inject_value"
]
